""""This library brings support for \
Yamaha MusicCast devices to Home Assistant."""
import queue
import socket
import logging
import threading
from homeassistant.const import (
    STATE_UNKNOWN, STATE_PLAYING, STATE_PAUSED, STATE_IDLE
)
from requests.exceptions import RequestException
from .const import ENDPOINTS
from .helpers import request, message_worker, socket_worker
from .media_status import MediaStatus
from .exceptions import YMCInitError
from .zone import Zone

_LOGGER = logging.getLogger(__name__)


class McDevice(object):
    """docstring for McDevice"""
    def __init__(self, ip_address, udp_port=5005, **kwargs):
        super(McDevice, self).__init__()
        _LOGGER.debug("McDevice: %s", ip_address)
        # construct message queue
        self.messages = queue.Queue()
        self._ip_address = ip_address
        self._udp_port = udp_port
        self._interval = kwargs.get('mc_interval', 480)
        self._zones = {}
        self._yamaha = None
        self._socket = None
        self._name = None
        self.device_id = None
        self.device_info = None
        self.device_features = None
        self.location_info = None
        self.update_status_timer = None
        try:
            self.initialize()
        except (OSError, RequestException) as err:
            raise YMCInitError(err)

    @property
    def ip_address(self):
        """Returns the ip_address."""
        return self._ip_address

    @property
    def zones(self):
        """Returns receiver zones."""
        return self._zones

    @zones.setter
    def zones(self, zones):
        """Sets receiver zones"""
        self._zones = zones

    @property
    def name(self):
        """Returns name of device."""
        return self._name

    @name.setter
    def name(self, name):
        """Sets name of device."""
        self._name = name

    def initialize(self):
        """initialize the object"""
        self.location_info = self.get_location_info()
        self.name = self.location_info.get('name', 'Unknown')
        self.device_info = self.get_device_info()
        self.device_id = (
            self.device_info.get('device_id')
            if self.device_info else "Unknown")
        self.initialize_socket()
        self.initialize_worker()
        self.initialize_zones()

    def initialize_socket(self):
        """initialize the socket"""
        self._socket = socket.socket(
            socket.AF_INET,     # IPv4
            socket.SOCK_DGRAM   # UDP
        )
        self._socket.bind(('', self._udp_port))
        _LOGGER.debug("Socket open.")
        _LOGGER.debug("Starting Socket Thread.")
        socket_thread = threading.Thread(
            name="SocketThread", target=socket_worker,
            args=(self._socket, self.messages,))
        socket_thread.setDaemon(True)
        socket_thread.start()

    def initialize_worker(self):
        """initialize the worker thread"""
        _LOGGER.debug("Starting Worker Thread.")
        worker_thread = threading.Thread(
            name="WorkerThread", target=message_worker, args=(self,))
        worker_thread.setDaemon(True)
        worker_thread.start()

    def initialize_zones(self):
        """initialize receiver zones"""
        zone_list = self.location_info.get('zone_list', {'main': True})

        for zone_id in zone_list:
            if zone_list[zone_id]:  # Location setup is valid
                self.zones[zone_id] = Zone(self, zone_id=zone_id)
            else:                   # Location setup is not valid
                _LOGGER.debug("Ignoring zone: %s", zone_id)

    def get_device_info(self):
        """Get info from device"""
        req_url = ENDPOINTS["getDeviceInfo"].format(self._ip_address)
        return request(req_url)

    def get_features(self):
        """Get features from device"""
        req_url = ENDPOINTS["getFeatures"].format(self._ip_address)
        return request(req_url)

    def get_location_info(self):
        """Get location info from device"""
        req_url = ENDPOINTS["getLocationInfo"].format(self._ip_address)
        return request(req_url)

    def get_status(self):
        """Get status from device"""
        headers = {
            "X-AppName": "MusicCast/0.1(python)",
            "X-AppPort": str(self._udp_port)
        }
        req_url = ENDPOINTS["getStatus"].format(self.ip_address, 'main')
        return request(req_url, headers=headers)

    def handle_netusb(self, message):
        """Handles 'netusb' in message"""
        # _LOGGER.debug("message: {}".format(message))
        if self._yamaha:
            if 'play_info_updated' in message:
                play_info = self.get_play_info()
                # _LOGGER.debug(play_info)
                if play_info:
                    self._yamaha.media_status = MediaStatus(
                        play_info, self._ip_address)
                    playback = play_info.get('playback')
                    # _LOGGER.debug("Playback: {}".format(playback))
                    if playback == "play":
                        self._yamaha.status = STATE_PLAYING
                    elif playback == "stop":
                        self._yamaha.status = STATE_IDLE
                    elif playback == "pause":
                        self._yamaha.status = STATE_PAUSED
                    else:
                        self._yamaha.status = STATE_UNKNOWN

    def handle_features(self, device_features):
        """Handles features of the device"""
        if device_features and 'zone' in device_features:
            for zone in device_features['zone']:
                zone_id = zone.get('id')
                if zone_id in self.zones:
                    _LOGGER.debug("handle_features: %s", zone_id)
                    input_list = zone.get('input_list', [])
                    input_list.sort()
                    self.zones[zone_id].source_list = input_list

    def handle_event(self, message):
        """Dispatch all event messages"""
        # _LOGGER.debug(message)
        for zone in self.zones:
            if zone in message:
                _LOGGER.debug("Received message for zone: %s", zone)
                self.zones[zone].handle_message(message[zone])

        if 'netusb' in message:
            self.handle_netusb(message['netusb'])

        self.update_hass()

    def update_status(self):
        """Update device status"""

        if not self.update_status_timer:
            _LOGGER.debug("update_status: First update")
            # try to get first device status, register for UDP Events
            status = self.get_status()
            # on success, schedule first timer
            if status:
                self.setup_update_timer()

                if not self.device_features:
                    # get device features only once
                    self.device_features = self.get_features()
                    self.handle_features(self.device_features)
                    self.update_hass()
        else:
            if not self.update_status_timer.is_alive():
                # e.g. computer was suspended, while hass was running
                _LOGGER.debug("update_status: Reschedule timer")
                self.setup_update_timer()

    def setup_update_timer(self):
        """Schedule a Timer Thread."""
        _LOGGER.debug(
            "update status: firing again in %d seconds", self._interval)
        self.update_status_timer = threading.Timer(
            self._interval, self.get_status)
        self.update_status_timer.setDaemon(True)
        self.update_status_timer.start()

    def update_hass(self):
        """Update HASS."""
        _LOGGER.debug("update_hass: Push updates")
        if self._yamaha and self._yamaha.entity_id:     # Push updates
            self._yamaha.schedule_update_ha_state()

    def set_yamaha_device(self, yamaha_device):
        """Set reference to device in HASS"""
        _LOGGER.debug("setYamahaDevice: %s", yamaha_device)
        self._yamaha = yamaha_device

    def get_play_info(self):
        """Get play info from device"""
        req_url = ENDPOINTS["getPlayInfo"].format(self._ip_address)
        return request(req_url)

    def set_playback(self, playback):
        """Send Playback command."""
        req_url = ENDPOINTS["setPlayback"].format(self._ip_address)
        params = {"playback": playback}
        return request(req_url, params=params)

    def __del__(self):
        if self._socket:
            _LOGGER.debug("Closing Socket.")
            self._socket.close()



#!/usr/bin/env python
"""This is a docstring."""
import logging
from homeassistant.const import (
    STATE_ON, STATE_OFF
)
from .const import ENDPOINTS
from .helpers import request
_LOGGER = logging.getLogger(__name__)


class Zone(object):
    """docstring for Zone"""
    def __init__(self, receiver, zone_id='main'):
        super(Zone, self).__init__()
        self._status = None
        self._zone_id = zone_id
        self._receiver = receiver
        self._yamaha = None
        self._ip_address = self.receiver.ip_address

    @property
    def status(self):
        """Returns status."""
        return self._status

    @status.setter
    def status(self, stat):
        self._status = stat

    @property
    def zone_id(self):
        """Returns the zone_id."""
        return self._zone_id

    @property
    def receiver(self):
        """Returns the receiver."""
        return self._receiver

    @property
    def ip_address(self):
        """Returns the ip_address."""
        return self._ip_address

    @property
    def source_list(self):
        """Return source_list."""
        return self._yamaha.source_list

    @source_list.setter
    def source_list(self, source_list):
        """Sets source_list."""
        self._yamaha.source_list = source_list

    def handle_message(self, message):
        """Process UDP messages"""
        if self._yamaha:
            if 'power' in message:
                _LOGGER.debug("Power: %s", message.get('power'))
                self._yamaha.power = (
                    STATE_ON if message.get('power') == "on" else STATE_OFF)
            if 'input' in message:
                _LOGGER.debug("Input: %s", message.get('input'))
                self._yamaha._source = message.get('input')
            if 'volume' in message:
                volume = message.get('volume')

                if 'max_volume' in message:
                    volume_max = message.get('max_volume')
                else:
                    volume_max = self._yamaha.volume_max

                _LOGGER.debug("Volume: %d / Max: %d", volume, volume_max)

                self._yamaha.volume = volume / volume_max
                self._yamaha.volume_max = volume_max
            if 'mute' in message:
                _LOGGER.debug("Mute: %s", message.get('mute'))
                self._yamaha.mute = message.get('mute', False)
        else:
            _LOGGER.debug("No yamaha-obj found")

    def update_status(self):
        """Updates the zone status."""
        if not self.status:
            _LOGGER.debug("update_status: Zone %s", self.zone_id)
            self.status = self.get_status()
            self.handle_message(self.status)
            self.update_hass()

    def update_hass(self):
        """Update HASS."""
        _LOGGER.debug("update_hass: Push updates")
        if self._yamaha and self._yamaha.entity_id:     # Push updates
            self._yamaha.schedule_update_ha_state()

    def get_status(self):
        """Get status from device"""
        req_url = ENDPOINTS["getStatus"].format(self.ip_address, self.zone_id)
        return request(req_url)

    def set_yamaha_device(self, yamaha_device):
        """Set reference to device in HASS"""
        _LOGGER.debug("setYamahaDevice: %s", yamaha_device)
        self._yamaha = yamaha_device

    def set_power(self, power):
        """Send Power command."""
        req_url = ENDPOINTS["setPower"].format(self.ip_address, self.zone_id)
        params = {"power": "on" if power else "standby"}
        return request(req_url, params=params)

    def set_mute(self, mute):
        """Send mute command."""
        req_url = ENDPOINTS["setMute"].format(self.ip_address, self.zone_id)
        params = {"enable": "true" if mute else "false"}
        return request(req_url, params=params)

    def set_volume(self, volume):
        """Send Volume command."""
        req_url = ENDPOINTS["setVolume"].format(self.ip_address, self.zone_id)
        params = {"volume": int(volume)}
        return request(req_url, params=params)

    def set_input(self, input_id):
        """Send Input command."""
        req_url = ENDPOINTS["setInput"].format(self.ip_address, self.zone_id)
        params = {"input": input_id}
        return request(req_url, params=params)



#!/usr/bin/env python
"""This file holds helper functions."""
import json
import time
import logging
import requests
_LOGGER = logging.getLogger(__name__)


def request(url, *args, **kwargs):
    """Do the HTTP Request and return data"""
    method = kwargs.get('method', 'GET')
    timeout = kwargs.pop('timeout', 10)  # hass default timeout
    req = requests.request(method, url, *args, timeout=timeout, **kwargs)
    data = req.json()
    _LOGGER.debug(json.dumps(data))
    return data


def message_worker(device):
    """Loop through messages and pass them on to right device"""
    msg_q = device.messages

    while True:

        if not msg_q.empty():
            message = msg_q.get()

            data = {}
            try:
                data = json.loads(message.decode("utf-8"))
            except ValueError:
                _LOGGER.error("Received invalid message: %s", message)

            if 'device_id' in data:
                device_id = data.get('device_id')
                if device_id == device.device_id:
                    device.handle_event(data)
                else:
                    _LOGGER.warning("Received message for unknown device.")
            msg_q.task_done()

        time.sleep(0.2)


def socket_worker(sock, msg_q):
    """Socket Loop that fills message queue"""
    while True:
        try:
            data, addr = sock.recvfrom(1024)    # buffer size is 1024 bytes
        except OSError as err:
            _LOGGER.error(err)
        else:
            _LOGGER.debug("received message: %s from %s", data, addr)
            msg_q.put(data)
        time.sleep(0.2)



#!/usr/bin/env python
"""This file defines the MediaStatus object."""


class MediaStatus(object):
    """docstring for MediaStatus"""
    def __init__(self, data, host):
        super(MediaStatus, self).__init__()
        self.host = host
        self.play_time = 0
        self.total_time = 0
        self.artist = None
        self.album = None
        self.track = None
        self.albumart_url = None
        self.initialize(data)

    @property
    def media_duration(self):
        """Duration of current playing media in seconds."""
        return self.total_time

    @property
    def media_image_url(self):
        """Image url of current playing media."""
        return "http://{}{}".format(self.host, self.albumart_url)

    @property
    def media_artist(self):
        """Artist of current playing media, music track only."""
        return self.artist

    @property
    def media_album(self):
        """Album of current playing media, music track only."""
        return self.album

    @property
    def media_track(self):
        """Track number of current playing media, music track only."""
        return self.track

    @property
    def media_title(self):
        """Title of current playing media."""
        return self.media_track

    def initialize(self, data):
        """ initialize variable from loaded data """
        for item in data:
            if hasattr(self, item):
                setattr(self, item, data[item])



#!/usr/bin/env python
"""This file defines constants."""

ENDPOINTS = {
    "getDeviceInfo": "http://{}/YamahaExtendedControl/v1/system/getDeviceInfo",
    "getFeatures": "http://{}/YamahaExtendedControl/v1/system/getFeatures",
    "getLocationInfo": ("http://{}/YamahaExtendedControl"
                        "/v1/system/getLocationInfo"),
    "getPlayInfo": "http://{}/YamahaExtendedControl/v1/netusb/getPlayInfo",
    "setPlayback": "http://{}/YamahaExtendedControl/v1/netusb/setPlayback",
    "getStatus": "http://{}/YamahaExtendedControl/v1/{}/getStatus",
    "setInput": "http://{}/YamahaExtendedControl/v1/{}/setInput",
    "setMute": "http://{}/YamahaExtendedControl/v1/{}/setMute",
    "setPower": "http://{}/YamahaExtendedControl/v1/{}/setPower",
    "setVolume": "http://{}/YamahaExtendedControl/v1/{}/setVolume",
}
